package edu.txstate.mag262;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Scanner;

/** A useful representation of a Princeton-Shape-Benchmark-style .cla file. */
public class CLAList {
	
	private File file;
	private double version;
	private ArrayList<CLAClass> classes;
	
	/** Parses information from a Princeton-Shape-Benchmark-style
	 *  .cla file to instantiate this object.
	 *  
	 * @param file	File object of .cla file.
	 */
	public CLAList(File file) {
		this.file = file;
		init();
	}
	
	/** Copy constructor
	 * 
	 * @param o	other CLAList object.
	 */
	public CLAList(CLAList o) {
		this.file = o.file;
		this.version = o.version;
		this.classes = new ArrayList<>();
		for (CLAClass ocla : o.classes) {
			this.classes.add(new CLAClass(ocla));
		}
	}
	
	/** Reads in a .cla file. */
	private void init() {
		this.classes = new ArrayList<>();
		
		try (Scanner s = new Scanner(new FileInputStream(file))) {
			s.next(); // Skip 'PSB' string.
			
			String t = s.next();
			if (t.matches("[0-9]+(\\.[0-9]+)*")) version = Double.parseDouble(t);
			else {
				System.err.println("Error: .cla file is improperly formatted.");
				System.err.println("\tProblem reading version number.");
				System.exit(1);
			}
			
			if (version > 1.0) {
				System.err.println("Error: only supports .cla file version 1.0.");
				System.exit(1);
			}
			
			int numberClasses = s.nextInt();
			int numberModels = s.nextInt();
			
			while (s.hasNext()) {
				String className = s.next();
				String parentClassName = s.next();
				int numClassModels = s.nextInt();
				ArrayList<Integer> models = new ArrayList<>();
				for (int i = 0; i < numClassModels; ++i)
					models.add(s.nextInt());
				classes.add(new CLAClass(className, parentClassName, models));
			}
			
			if (classes.size() != numberClasses) {
				System.err.println("Error: mismatch between number of classes listed and found in .cla file.");
				System.exit(1);
			}
			
			if (getNumberModels() != numberModels) {
				System.err.println("Error: mismatch between number of models listed and found in .cla file.");
				System.exit(1);
			}
				
		} catch (IOException e) {
			System.err.println("Error reading .cla file in CLAList instantiation.");
		}
	}
	
	/** Saves CLA information at a given filename in PSB Classification File Format.
	 * 
	 * @param filename	path to save to (should end in .cla).
	 */
	public void toFile(String filename) {
		try (BufferedWriter w = new BufferedWriter(new FileWriter(filename))) {
			w.write("PSB 1\n");
			w.write(classes.size() + " " + getNumberModels() + "\n\n");
			
			for (CLAClass c : classes) {
				ArrayList<Integer> mList = c.getModelList();
				w.write(c.getName() + " " + c.getParent() + " " + mList.size() + "\n");
				for (int i : mList) w.write(i + "\n");
				w.write("\n");
			}
			
		} catch (IOException e) {
			System.err.println("Error: could not write .cla file.");
		}
	}
	
	/** @return A pointer to the internal .cla File object used. */
	public File getFile() { return file; }
	
	/** @return The version number of the .cla file format. */
	public double getVersion() { return version; }

	/** @return The number of classes listed in the .cla file. */
	public int getNumberClasses() { return classes.size(); }

	/** @return The total number of models listed in the .cla file. */
	public int getNumberModels() { 
		int numModels = 0;
		for (CLAClass c : classes) numModels += c.numModels();
		return numModels;
	}

	/** @return A pointer to the list of CLAClass objects of each class.
	 * @see CLAClass */
	public ArrayList<CLAClass> getClasses() { return classes; }

	@Override
	public String toString() {
		String s = "Version " + version + "\n";
		s += getNumberClasses() + " classes\n";
		s += getNumberModels() + " models\n";
		return s;
	}
}
