package edu.txstate.mag262;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Scanner;

/** The global histogram of a 3D model.*/
public class GlobalHistogram {
	private ShapeDescriptor shape;
	private double histogram[];
	
	/** Generates a global histogram of a Model3D object.
	 * 
	 * @param model		the Model3D object.
	 * @param shells	the number of shells to use in histogram. (>= 1)
	 * @param sphere	a SphereApproximator object used to describe the sectors.
	 * 
	 * @see SphereApproximator
	 */
	public GlobalHistogram(Model3D model, int shells, SphereApproximator sphere) {
		this.shape = new ShapeDescriptor(model.getBoundingSphere().getRadius(), shells, sphere);
		
		histogram = new double[shape.getNumPoints()];
		Arrays.fill(histogram, 0);
		
		ArrayList<Vertex3D> vertices = model.getVertices();
		ArrayList<Vertex3D> bins = shape.getPoints();
		
		// Assign model vertices to bins.
		for (Vertex3D v : vertices) {
			
			int c = 0; // Index of closest bin.
			double minSqDist = v.sqDistanceTo(bins.get(c));
			
			// Iterate through all bin points to find one closest to 'v'.
			for (int i = 1; i < shape.getNumPoints(); ++i) {
				double sqDist = v.sqDistanceTo(bins.get(i));
				if (sqDist < minSqDist) {
					minSqDist = sqDist;
					c = i;
				}
			}
			
			// Increment value at the closest bin.
			histogram[c] += 1;
		}
		
		// Finish building percentage histogram.
		for (int i = 0; i < shape.getNumPoints(); ++i) histogram[i] /= vertices.size();
	}
	
	/** Generates a global histogram of a Model3D object. 
	 * 
	 * @param model		the Model3D object.
	 * @param shells	the number of shells to use in histogram. (>= 1)
	 * @param order		number of subdivisions of sphere approximation.
	 * 
	 * @see SphereApproximator
	 * */
	public GlobalHistogram(Model3D model, int shells, int order) {
		this(model, shells, new SphereApproximator(order));
	}
	
	/** Generates a global histogram of a Model3D object. 
	 * 
	 * @param model			the Model3D object.
	 * @param shells		the number of shells to use in histogram. (>= 1)
	 * @param horizontal	number of horizontal sectors.
	 * @param vertical		number of vertical sectors.
	 * 
	 * @see SphereApproximator
	 * */
	public GlobalHistogram(Model3D model, int shells, int horizontal, int vertical) {
		this(model, shells, new SphereApproximator(horizontal , vertical));
	}
	
	/** Compares this global histogram with a generated directory of global histograms.
	 * Writes the results to a result file which will have the distance and model number
	 * per line.
	 * 
	 * @param histograms	the path to a generated directory of global histograms.
	 * @param resultFile	the path to the result file.
	 * @param similarity	at high values (above 10 or so) the distance function resembles
	 * 		more the basic Euclidean distance; at low values (around 1), the function 
	 * 		incorporates more information about neighboring histogram bins.
	 * 
	 * @see HistogramDirectory
	 * @see QuadraticFormDistance
	 * */
	public void compareWith(String histograms, String resultFile, double similarity) {
		
		System.out.print("Initializing similarity matrix...");
		QuadraticFormDistance qfd = new QuadraticFormDistance(shape.getPoints(), similarity);
		System.out.println("done.");
		
		ArrayList<Double> distances = new ArrayList<>();
		ArrayList<Integer> modelIDs = new ArrayList<>();
		
		System.out.print("Comparing model to database...");
		
		// Get file list.
		String[] subNodes = new File(histograms).list();
		for (String filename : subNodes) {
			
			// Get model number from histogram filename.
			int mID = 0;
			try {
				mID = Integer.parseInt(filename.replaceAll("[A-Za-z\\.]+", ""));
			} catch (NumberFormatException e) {
				System.err.println("Error: problem extracting model number from file!");
				System.err.println("\tFilename: " + filename);
			}
			
			// Get histogram data.
			double other[] = openFromFile(new File(histograms, filename));
			
			// Measure distance from other histogram to this one.
			double dist = qfd.getDistance(histogram, other);
			
			// Insert the error and model number in order.
			if (distances.isEmpty()) {
				distances.add(dist);
				modelIDs.add(mID);
			} else {
				boolean inserted = false;
				
				for (int i = 0; i < distances.size(); ++i) {
					if (distances.get(i) > dist) {
						distances.add(i, dist);
						modelIDs.add(i, mID);
						inserted = true;
						break;
					}
				}
				
				if (!inserted) {
					distances.add(dist);
					modelIDs.add(mID);
				}
			}
		}
		
		System.out.print("done.\nWriting results to "+resultFile+"...");
		
		// Write results to file.
		try (BufferedWriter writer = new BufferedWriter(new FileWriter(resultFile))) {
			
			for (int i = 0; i < distances.size(); ++i) 
				writer.write(String.format("%1.15f %d\n", distances.get(i), modelIDs.get(i)));
			
		} catch (IOException e) {
			System.out.println("Error creating results file!");
		}
		System.out.println("done.");
	}
	
	/** Returns the distance matrix of global histograms in a generated histograms directory.
	 * 
	 * @param histograms	the path to a generated directory of global histograms.
	 * @param claFile		the .cla file to use to generate the distance matrix.
	 * @param outputFile	the filepath where you want to save your results.
	 * @param shape			a global ShapeDescriptor with the same shape used in generating the
	 * 		histograms directory.
	 * @param similarity	at high values (above 10 or so) the distance function resembles
	 * 		more the basic Euclidean distance; at low values (around 1), the function 
	 * 		incorporates more information about neighboring histogram bins.
	 * 
	 * @see QuadraticFormDistance
	 * @see ShapeDescriptor
	 * */
	public static double[][] createDistanceMatrix(String histograms, String claFile, String outputFile, ShapeDescriptor shape, double similarity) {
		
		System.out.print("Initializing similarity matrix...");
		QuadraticFormDistance qfd = new QuadraticFormDistance(shape.getPoints(), similarity);
		System.out.println(" done.");
		
		double matrix[][] = null;
		
		try {
			CLAList cList = new CLAList(new File(claFile));
			
			// Get list of files in directory.
			File file = new File(histograms);
			if (!file.isDirectory()) {
				System.out.println("Hey, histogram directory given isn't a directory.");
				System.exit(1);
			} 
			File files[] = file.listFiles();
			
			// Load histograms.
			System.out.print("Loading histograms...");
			ArrayList<double[]> hists = new ArrayList<>();
			ArrayList<Integer> leftOut = new ArrayList<>();
			for (CLAClass cla : cList.getClasses()) {
				for (Integer m : cla.getModelList()) {
					boolean found = false;
					
					// Search all files for each model number.
					for (File f : files) {
						String name = f.getName();
						String mST = name.replaceAll("[A-Za-z\\.]+", "");
						if (!mST.matches("[0-9]+")) continue;
						int mID = Integer.parseInt(mST);
						
						// If corresponding histogram found, add it to list.
						if ((int)m == mID) {
							hists.add(openFromFile(new File(histograms, name)));
							found = true;
						}
					}
					
					// If model wasn't found in histogram directory...
					if (!found) leftOut.add(m);
				}
			}
			System.out.println("done.");
			
			// Check if models were left out (histogram didn't exist for it).
			if (leftOut.size() > 0) {
				System.out.printf("%d models in .cla file were not in histogram directory!\n", leftOut.size());
				for (CLAClass cla : cList.getClasses()) {
					for (Integer i : leftOut) {
						if (cla.getModelList().contains(i)) {
								cla.getModelList().remove(i);
						}
					}
				}
				System.out.println("New .cla file created for use with psbTable.exe (revised.cla).");
				cList.toFile("revised.cla");
			}
			
			if (hists.size() != cList.getNumberModels()) {
				System.err.println("Error: well, this is weird...check the code to createDistanceMatrix(s,s,s, sd, d) in GlobalHistogram.java");
				System.exit(1);
			}
			
			int numModels = hists.size();
			int numEntries = numModels*numModels;
			
			matrix = new double[numModels][numModels]; // Allocate matrix space.
			
			System.out.printf("Comparing %d models...\n", numModels);
			System.out.printf("\tMatrix has %d entries.\n", numEntries);
			
			// Compute distances down diagonal (should all be zero).
			for (int i = 0; i < numModels; ++i) matrix[i][i] = 0;
			
			// Compute distances between different models.
			System.out.print("\tCreating matrix...");
			int cnt = numModels;
			for (int i = 0; i < numModels; ++i) {
				for (int j = 0; j < i; ++j) {
					matrix[i][j] = matrix[j][i] = qfd.getDistance(hists.get(i), hists.get(j));
					cnt += 2;
				}
				System.out.printf("\r\tCreating matrix... %.3f%%    ", (cnt * 100.0) / numEntries);
			}
			System.out.println("\r\tCreating matrix... done.        ");
			System.out.println("done.");
			
			// Write matrix to file.
			BufferedWriter writer = new BufferedWriter(new FileWriter(outputFile));
			for (int i = 0; i < numModels; ++i) {
				for (int j = 0; j < numModels; ++j) writer.write(matrix[i][j] + " ");
				writer.write("\n");
			}
			writer.close();
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.out.println("Error reading model ID file!");
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("Error writing output file!");
		}
		
		return matrix;
	}
	
	/** Returns the distance matrix of histograms in a generated histograms directory. Uses
	 * Euclidian distance for comparison. Format of histogram file should be just the value
	 * of each bin per line.
	 * 
	 * @param histograms	the path to a generated directory of global histograms.
	 * @param claFile		the .cla file to use to generate the distance matrix.
	 * @param outputFile	the filepath where you want to save your results.
	 * */
	public static double[][] createDistanceMatrix(String histograms, String claFile, String outputFile) {
		
		double[][] matrix = null;
		
		try {
			CLAList cList = new CLAList(new File(claFile));
			
			// Read in a model to get information.
			File file = new File(histograms);
			int dimension = 0;
			String filetype = "";
			String prepend = "";
			File[] files = file.listFiles();
			
			if (file.isDirectory()) {
				for (int i = 0; i < files.length; ++i) {
					
					// If the filename looks like it's a model file.
					if (files[i].getName().matches("[a-zA-z]+[0-9]+\\.[a-zA-Z]+")) {
						
						// Get information about how model files are named.
						filetype = files[i].getName().replaceAll("[a-zA-z]+[0-9]+\\.", "");
						prepend = files[i].getName().replaceAll("[0-9]+\\."+filetype, "");
						
						// Read file to get dimension of histograms.
						Scanner scanner = new Scanner(files[i]);
						while (scanner.hasNextDouble()) {
							++dimension;
							scanner.nextDouble();
						}
						scanner.close();
						break;
					}
				}
			} else {
				System.out.println("Hey, histogram directory given isn't a directory.");
				System.exit(1);
			}
			
			System.out.println("Dimension: " + dimension);
			System.out.println("Filetype: " + filetype);
			System.out.println("Prepend: " + prepend);
			
			// Load histograms.
			System.out.print("Loading histograms...");
			ArrayList<double[]> hists = new ArrayList<>();
			ArrayList<Integer> leftOut = new ArrayList<>();
			for (CLAClass cla : cList.getClasses()) {
				for (Integer m : cla.getModelList()) {
					boolean found = false;
					
					// Search all files for each model number.
					for (File f : files) {
						String name = f.getName();
						String mST = name.replaceAll("[A-Za-z\\.]+", "");
						if (!mST.matches("[0-9]+")) continue;
						int mID = Integer.parseInt(mST);
						
						// If corresponding histogram found, add it to list.
						if ((int)m == mID) {
							hists.add(openFromFileStraight(new File(histograms, name)));
							found = true;
						}
					}
					
					// If model wasn't found in histogram directory...
					if (!found) leftOut.add(m);
				}
			}
			System.out.println("done.");
			
			// Check if models were left out (histogram didn't exist for it).
			if (leftOut.size() > 0) {
				System.out.printf("%d models in .cla file were not in histogram directory!\n", leftOut.size());
				for (CLAClass cla : cList.getClasses()) {
					for (Integer i : leftOut) {
						if (cla.getModelList().contains(i)) {
								cla.getModelList().remove(i);
						}
					}
				}
				System.out.println("New .cla file created for use with psbTable.exe (revised.cla).");
				cList.toFile("revised.cla");
			}
			
			if (hists.size() != cList.getNumberModels()) {
				System.err.println("Error: well, this is weird...check the code to createDistanceMatrix(s,s,s) in GlobalHistogram.java");
				System.exit(1);
			}
			
			int numModels = hists.size();
			int numEntries = numModels*numModels;
			
			matrix = new double[numModels][numModels]; // Allocate matrix space.
			
			System.out.printf("Comparing %d models...\n", numModels);
			System.out.printf("\tMatrix has %d entries.\n", numEntries);
			
			// Compute distances down diagonal (should all be zero).
			for (int i = 0; i < numModels; ++i) matrix[i][i] = 0;
			
			// Compute distances between different models.
			System.out.print("\tCreating matrix...");
			int cnt = numModels;
			for (int i = 0; i < numModels; ++i) {
				for (int j = 0; j < i; ++j) {
					matrix[i][j] = matrix[j][i] = Vertex3D.sqDistance(hists.get(i), hists.get(j));
					cnt += 2;
				}
				System.out.printf("\r\tCreating matrix... %.3f%%    ", (cnt * 100.0) / numEntries);
			}
			
			System.out.println("\r\tCreating matrix... done.        ");
			System.out.println("done.");
			
			// Write histogram to file.
			BufferedWriter writer = new BufferedWriter(new FileWriter(outputFile));
			for (int i = 0; i < numModels; ++i) {
				for (int j = 0; j < numModels; ++j) writer.write(matrix[i][j] + " ");
				writer.write("\n");
			}
			writer.close();
			
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("Error opening a file!");
		}
		
		return matrix;
	}
	
	/** Save as histogram data to file with given filename.
	 * <br><br>
	 * Format of the saved file is as follows:
	 * <ol>
	 * <li>First line contains the number of bins in the histogram.</li>
	 * <li>Following lines each hold the index of a bin followed by it's non-zero value
	 * (separated by a space).</li>
	 * <li>Zero-valued bins are excluded.</li>
	 * </ol>
	 * 
	 * @param filename	filename of histogram, "something.ghist"
	 */
	public void saveToFile(String filename) {
		try (BufferedWriter writer = new BufferedWriter(new FileWriter(filename))){
			writer.write(shape.getNumPoints() + "\n");
			for (int i = 0; i < shape.getNumPoints(); ++i)
				if (!Vertex3D.equalEps(histogram[i], 0))
					writer.write(String.format("%d %.18f\n", i, histogram[i]));
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error creating percent histogram file.");
		}
	}
	
	/** Load a previously saved global histogram vector.
	 * 
	 * @param file	a File object of the histogram file.
	 * @return percent histogram vector in a double array.
	 */
	public static double[] openFromFile(File file) {
		double other[] = null;
		try (Scanner scanner = new Scanner(new FileInputStream(file))) {
			other = new double[scanner.nextInt()];
			Arrays.fill(other, 0);
			while (scanner.hasNextInt())
				other[scanner.nextInt()] = scanner.nextDouble();
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error opening histogram file!");
		}
		return other;
	}
	
	/** Load a previously saved histogram vector. One double per line.
	 * No special formatting.
	 * 
	 * @param file	a File object of the histogram file.
	 * @return percent histogram vector in a double array.
	 */
	public static double[] openFromFileStraight(File file) {
		ArrayList<Double> other = new ArrayList<>();
		try (Scanner scanner = new Scanner(new FileInputStream(file))) {
			while (scanner.hasNextDouble()) other.add(scanner.nextDouble());
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error opening histogram file!");
		}
		double[] result = new double[other.size()];
		for (int i = 0; i < other.size(); ++i) result[i] = other.get(i);
		return result;
	}

}