package edu.txstate.mag262;
import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

/** Used to create a directory of global histograms from a model database. */
public class HistogramDirectory {
	/** Normalize the models from scratch against rotations and scale. */
	public static final int NORMALIZE = 0;
	/** Use model_info files for normalizing the models against rotations. */
	public static final int MODEL_INFO = 1;
	/** Perform no normalizations on the models. */
	public static final int VANILLA = 2;
	
	private int shells;
	private SphereApproximator sphere;
	private Path storagePath;
	private Path modelPath;
	
	@SuppressWarnings("unused")
	private HistogramDirectory() {}
	
	/** Constructs object with the information necessary to create a directory of global histograms 
	 * from a model database.
	 * 
	 * @param modelPath		path to a directory containing models. Could be a directory of directories.
	 * @param storagePath	path to the storage directory.
	 * @param shells		number of shells to use in global histogram.
	 * @param sphere		used to determine the number sectors to use in global histogram.
	 * 
	 * @see SphereApproximator
	 */
	public HistogramDirectory(String modelPath, String storagePath, int shells, SphereApproximator sphere) {
		this.modelPath = Paths.get(modelPath);
		this.storagePath = Paths.get(storagePath);
		this.shells = shells;
		this.sphere = sphere;
	}
	
	/** Generates a global histogram file ".ghist" for each model found. Histograms are 
	 * stored in a single directory.
	 *  
	 *  @param mode		use HistogramDirectory.(NORMALIZE|MODEL_INFO|VANILLA)
	 *  */
	public void generate(int mode) {
		
		// Try creating a new directory with the given storage Path (for storing the generated histograms).
		try { Files.createDirectory(storagePath);
		
		// If that fails, directory probably already exists.
		} catch (Exception e) {}
		
		// Generate histograms.
		switch (mode) {
			case NORMALIZE: generateAuxNormalize(modelPath.toFile()); break;
			case MODEL_INFO: generateAuxInfo(modelPath.toFile()); break;
			case VANILLA: generateAuxVanilla(modelPath.toFile()); break;
		}
		
		System.out.println("\rFinished generating histograms!                  ");
	}
	
	/** Recursively finds model files, normalizes against several factors, and generates histograms. */
	private void generateAuxNormalize(File node) {
		
		// If node is a directory, get file list.
		if(node.isDirectory()) {
			
			String[] subNodes = node.list();
			for (String filename : subNodes)
				generateAuxNormalize(new File(node, filename));
			
	    // If node is a model file, generate histogram file.
		} else if (node.getName().toLowerCase().matches("[\\s\\S]+((\\.of{2})|(\\.obj))")) {
			
			Model3D model = new Model3D(node.getAbsolutePath(), true);
			GlobalHistogram hist = new GlobalHistogram(model, shells, sphere);
			
			hist.saveToFile(storagePath.toAbsolutePath() + System.getProperty("file.separator") + node.getName().replaceAll("(off)|(obj)", "ghist"));
			System.out.print("\rGenerating histogram " + node.getName().replaceAll("(\\.off)|(\\.obj)", "") + "       ");
		}
	}
	
	/** Recursively finds model files AND uses their model info files. Assumes models names
	 * follow the Princeton Shape Benchmark convention "mX.off" and have a model info
	 * file in the same directory "mX_info.txt", where X is a number. */
	private void generateAuxInfo(File node) {
		
		// If node is a directory, get file list.
		if(node.isDirectory()){
			
			String[] subNodes = node.list();
			for (String filename : subNodes)
				generateAuxInfo(new File(node, filename));
			
	    // If node is a model file, generate histogram file.
		} else if (node.getName().toLowerCase().matches("[\\s\\S]+(\\.of{2})")) {
			String filepath = node.getAbsolutePath();
			
			Model3D model = new Model3D(filepath, filepath.replace(".off", "_info.txt"));
			GlobalHistogram hist = new GlobalHistogram(model, shells, sphere);
			
			hist.saveToFile(storagePath.toAbsolutePath() + System.getProperty("file.separator") + node.getName().replaceAll("(off)|(obj)", "ghist"));
			System.out.print("\rGenerating histogram " + node.getName().replaceAll("(\\.off)|(\\.obj)", "") + "       ");
		}
	}
	
	/** Recursively finds model files and generates histograms straight from model data.  */
	private void generateAuxVanilla(File node) {
		
		// If node is a directory, get file list.
		if(node.isDirectory()){
			
			String[] subNodes = node.list();
			for (String filename : subNodes)
				generateAuxVanilla(new File(node, filename));
			
	    // If node is a model file, generate histogram file.
		} else if (node.getName().toLowerCase().matches("[\\s\\S]+((\\.of{2})|(\\.obj))")) {
			String filepath = node.getAbsolutePath();
			
			Model3D model = new Model3D(filepath, false);
			GlobalHistogram hist = new GlobalHistogram(model, shells, sphere);
			
			hist.saveToFile(storagePath.toAbsolutePath() + System.getProperty("file.separator") + node.getName().replaceAll("(off)|(obj)", "ghist"));
			System.out.print("\rGenerating histogram " + node.getName().replaceAll("(\\.off)|(\\.obj)", "") + "       ");
		}
	}
}
