package edu.txstate.mag262;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.StringTokenizer;

import Jama.EigenvalueDecomposition;
import Jama.Matrix;

/** A collection of vertices and bounding sphere of a 3D model. */
public class Model3D {
	
	private ArrayList<Vertex3D> vertices;
	private BoundingSphere bsphere;
	
	@SuppressWarnings("unused")
	private Model3D() {}
	
	/** Reads in the vertices of a 3D model and performs (optional) normalization
	 * on them before calculating the bounding sphere. Normalization means
	 * applying PCA and scaling isotropically by mean distance of points.
	 *  
	 *  @param filename		the path to the 3D model file. Only .off and .obj files are
	 *  currently supported.
	 *  @param normalize	if true, normalizes the model against various factors; else
	 *  it just uses the model as-is.
	 *   */
	public Model3D(String filename, boolean normalize)  {
		vertices = new ArrayList<>();
		
		try {
			readPointsFromFile(filename, vertices);
		} catch (IOException e) {
			System.err.println("Error reading from object file in constructing Model3D object.");
		}
		
		if (normalize) {
			subtractMeans(vertices);
			rescaleByMean(1.5, vertices);
			applyPcaTransform(vertices);
		}
		
		bsphere = BoundingSphere.boundingBox(vertices);
		bsphere.toOrigin(vertices);
	}
	
	/** Reads in the vertices of a 3D model and performs some normalization
	 * on them before calculating the bounding sphere. Normalization means
	 * applying PCA and scaling isotropically by mean distance of points.
	 * In this constructor, the PCA basis vectors are extracted from a 
	 * Princeton Shape Benchmark style model_info file.
	 *  
	 *  @param filename		the path to the 3D model file. Only .off and .obj files are
	 *  	currently supported.
	 *  @param modelInfo	the PSB style model_info file.
	 *  */
	public Model3D(String filename, String modelInfo)  {
		vertices = new ArrayList<>();
		try {
			readPointsFromFile(filename, vertices);
		} catch (IOException e) {
			System.err.println("Error reading from object file in constructing Model3D object.");
		}
		ModelInfo info = readModelInfo(modelInfo);
		subtractMeans(vertices);
		rescaleByMean(1.5, vertices);
		applyPcaTransform(vertices, info.principalAxes);
		bsphere = BoundingSphere.boundingBox(vertices);
		bsphere.toOrigin(vertices);
	}
	
	/** @return Deep copy of the model's vertices. */
	public ArrayList<Vertex3D> getVertices() { 
		ArrayList<Vertex3D> copy = new ArrayList<>();
		for (Vertex3D v : vertices) copy.add(new Vertex3D(v));
		return copy;
	}
	
	/** @return The associated BoundingSphere object. */
	public BoundingSphere getBoundingSphere() { return bsphere; }

	/** Finds furthest point from the origin. 
	 * 
	 * @param points	a list of points in 3D space.
	 * @return The distance from the origin to the furthest point.
	 */
	public static double furthestPoint(ArrayList<Vertex3D> points) {
		double max = points.get(0).length();
		for (int i = 1; i < points.size(); ++i) {
			double h = points.get(i).length();
			if (h > max) max = h;
		}
		return max;
	}
	
	/** Rescales a set of points in 3D space in respect to the origin. The furthest point
	 * from the origin will be at the given radius after scaling.
	 * 
	 * @param radius	the new radius.
	 * @param points	a list of points in 3D space.
	 */
	public static void rescale(double radius, ArrayList<Vertex3D> points) {
		double max = furthestPoint(points);
		for (Vertex3D v : points) v.scaleToLength((v.length() / max) * radius);
	}
	
	/** Rescales a set of points in 3D space in respect to the origin. The mean of
	 * distances from the origin will be at the given radius after scaling.
	 * 
	 * @param radius	the new radius.
	 * @param points	a list of points in 3D space.
	 */
	public static void rescaleByMean(double radius, ArrayList<Vertex3D> points) {
		double meanLength = 0;
		for (Vertex3D v : points) meanLength += v.length();
		meanLength /= points.size();
		for (Vertex3D v : points) v.scaleToLength(v.length() * radius / meanLength);
	}
	
	/** Subtracts the means from each dimension so that each dimension has mean of zero.
	 * 
	 * @param points	a list of points in 3D space.
	 */
	public static void subtractMeans(ArrayList<Vertex3D> points) {
		Vertex3D mean = new Vertex3D(0, 0, 0);
		for (Vertex3D v : points) mean.add(v);
		mean.times(1.0 / points.size());
		for (Vertex3D v : points) v.sub(mean);
	}
	
	/** Rotates a set points so that they're aligned with the pricipal components of
	 * the set.
	 * 
	 * @param points	a list of points in 3D space.
	 */
	public static void applyPcaTransform(ArrayList<Vertex3D> points) {
		double cov_matrix[][] = new double[3][3];
		
		// Make sure covariance matrix is zeroed out.
		for (int i = 0; i < 3; ++i) Arrays.fill(cov_matrix[i], 0);
		
		// Create covariance matrix.
		double n = 1.0 / (points.size() - 1);
		for (int i = 0; i < points.size(); ++i) {
			cov_matrix[0][0] += points.get(i).x * points.get(i).x * n;
			cov_matrix[0][1] += points.get(i).x * points.get(i).y * n;
			cov_matrix[0][2] += points.get(i).x * points.get(i).z * n;
			
			cov_matrix[1][0] += points.get(i).y * points.get(i).x * n;
			cov_matrix[1][1] += points.get(i).y * points.get(i).y * n;
			cov_matrix[1][2] += points.get(i).y * points.get(i).z * n;
			
			cov_matrix[2][0] += points.get(i).z * points.get(i).x * n;
			cov_matrix[2][1] += points.get(i).z * points.get(i).y * n;
			cov_matrix[2][2] += points.get(i).z * points.get(i).z * n;
		}
		
		// Get eigenvector matrix (Using JAMA library).
		Matrix cov = new Matrix(cov_matrix);
		EigenvalueDecomposition eig_dec = cov.eig();
		Matrix eigvec_matrix = new Matrix(3, 3);
		double eig_values[] = eig_dec.getRealEigenvalues();
		
		// Sort eigenvectors in matrix according to eigenvalues (high -> low).
		int count = 0;
		while (count < 3) {
			double max_v = 0;
			int max_i = 0;
			for (int i = 0; i < 3; ++i) {
				if (eig_values[i] > max_v) {
					max_v = eig_values[i];
					max_i = i;
				}
			}
			for (int i = 0; i < 3; ++i)
				eigvec_matrix.set(i, count, eig_dec.getV().get(i, max_i));
			max_v = 0;
			eig_values[max_i] = 0;
			++count;
		}
		
		// Make a matrix from the model vertices, the rows are x y z, columns are samples.
		Matrix vert = new Matrix(3, points.size());
		for (int i = 0; i < points.size(); ++i) {
			vert.set(0, i, points.get(i).x);
			vert.set(1, i, points.get(i).y);
			vert.set(2, i, points.get(i).z);
		}
		
		// Transform the vertices in respect to the principal axes.
		Matrix new_vert = (eigvec_matrix.transpose()).times(vert);
		
		// Make vertices the new points.
		for (int i = 0; i < points.size(); ++i) {
			points.get(i).x = new_vert.get(0, i);
			points.get(i).y = new_vert.get(1, i);
			points.get(i).z = new_vert.get(2, i);
		}
	}
	
	/** Rotates a set points so that they're aligned with the pricipal components of
	 * the set, given a PCA basis. 
	 * 
	 * @param points	a list of points in 3D space.
	 * @param pcaBasis	a matrix of where the columns are three dimensional PCA 
	 * 	basis vectors. For example, pcaBasis[0][2] will be row 0, column 2; using
	 * 	zero-indexing.
	 * */
	public static void applyPcaTransform(ArrayList<Vertex3D> points, double[][] pcaBasis) {
		Matrix eigvec_matrix = new Matrix(pcaBasis, 3, 3);
		
		// Make a matrix from the model vertices, the rows are x y z, columns are samples.
		Matrix vert = new Matrix(3, points.size());
		for (int i = 0; i < points.size(); ++i) {
			vert.set(0, i, points.get(i).x);
			vert.set(1, i, points.get(i).y);
			vert.set(2, i, points.get(i).z);
		}
		
		// Transform the vertices in respect to the principal axes.
		Matrix new_vert = (eigvec_matrix.transpose()).times(vert);
		
		// Make vertices the new points.
		for (int i = 0; i < points.size(); ++i) {
			points.get(i).x = new_vert.get(0, i);
			points.get(i).y = new_vert.get(1, i);
			points.get(i).z = new_vert.get(2, i);
		}
	}
	
	/** Reads a Princeton Shape Benchmark style model_info file.
	 *  
	 * @param filename	the model info filepath.
	 * 
	 * @return a ModelInfo object with the extracted model information.
	 * 
	 * @see Model3D.ModelInfo
	 */
	public static ModelInfo readModelInfo(String filename) {
		ModelInfo info = new ModelInfo();
		
		try (BufferedReader reader = new BufferedReader(new FileReader(filename))){
			String notNum = "[^\\-?\\.\\d]+";
			String num = "[\\-?\\.\\d]+";
			String s = reader.readLine();
			while (s != null) {
				String tokens[] = s.split(notNum); // Split line string.
				
				// Find token with first number.
				int firstNum = 0;
				for (int i = 0; i < tokens.length; ++i) {
					if (tokens[i].matches(num)) {
						firstNum = i;
						break;
					}
				}
				
				if (s.contains("center")) {
					info.center = new Vertex3D(Double.parseDouble(tokens[firstNum++]),
											   Double.parseDouble(tokens[firstNum++]),
										   	   Double.parseDouble(tokens[firstNum++]));
				} else if (s.contains("scale")) {
					info.scale = Double.parseDouble(tokens[firstNum]);
				} else if (s.contains("principle_axes")) {
					info.principalAxes = new double[][] {
							{Double.parseDouble(tokens[firstNum]), Double.parseDouble(tokens[firstNum+3]), Double.parseDouble(tokens[firstNum+6])},
							{Double.parseDouble(tokens[firstNum+1]), Double.parseDouble(tokens[firstNum+4]), Double.parseDouble(tokens[firstNum+7])},
							{Double.parseDouble(tokens[firstNum+2]), Double.parseDouble(tokens[firstNum+5]), Double.parseDouble(tokens[firstNum+8])}
					};
				} else if (s.contains("principle_values")) {
					info.principalVales = new double[] {Double.parseDouble(tokens[firstNum++]), Double.parseDouble(tokens[firstNum++]), Double.parseDouble(tokens[firstNum++])};
				}
				s = reader.readLine();
			}
			
		} catch (FileNotFoundException e) {
			System.out.println("Model info file not found!");
		} catch (IOException e1) {
			System.out.println("Error closing model info file!");
		}
		
//		System.out.println("Center: " + info.center);
//		System.out.println("Scale: " + info.scale);
//		System.out.println("Principal axes: ");
//		for (int i = 0; i < 3; ++i) {
//			System.out.print("\t");
//			for (int j = 0; j < 3; ++j) {
//				System.out.print(info.principalAxes[i][j] + " ");
//			}
//			System.out.println();
//		}
//		System.out.print("Principal values: ");
//		for (int i = 0; i < 3; ++i) System.out.print(info.principalVales[i] + " ");
//		System.out.println();
		
		return info;
	}
	
	/** Reads in a model file and returns the vertices as an ArrayList of Vertex3D objects.
	 *  
	 * @param filename	the path to the 3D model file. Only .off and .obj files are
	 *  	currently supported.
	 * @param vertices	an instantiated ArrayList to add the model vertices to.
	 * 
	 * @throws IOException if model file could not be opened.
	 */
	public static void readPointsFromFile(String filename, ArrayList<Vertex3D> vertices) throws IOException {
		// Read .off file.
		if (filename.toLowerCase().endsWith("off")) {
			
			StreamTokenizer reader = new StreamTokenizer(new FileReader(filename));
		
			reader.nextToken(); // First token should be the string "OFF"
			reader.nextToken();
			
			int numVertices = (int)reader.nval; // Next is number of vertices.
			
			// Don't need next two tokens.
			reader.nextToken();
			reader.nextToken();
			
			// Read the vertex data.
			double x, y, z;
			for (int i = 0; i < numVertices; ++i) {
				
				// Read x.
				reader.nextToken();
				x = reader.nval;
				
				// Read y.
				reader.nextToken();
				y = reader.nval;
				
				// Read z.
				reader.nextToken();
				z = reader.nval;
				
				// Make vertex and add to Model3D data.
				vertices.add(new Vertex3D(x, y, z));
			}
		
		// Or ".obj" file.
		} else if (filename.toLowerCase().endsWith("obj")) {
			
			BufferedReader reader = new BufferedReader(new FileReader(filename));
			
			String line = reader.readLine();
			StringTokenizer tokenizer = new StringTokenizer(line);
			
			// Read the vertex data.
			while (line != null && tokenizer.countTokens() == 4) {
				tokenizer.nextToken(); // Skip "v" character.
				
				// Make vertex and add to Model3D data.
				Vertex3D vertex = new Vertex3D(
						Double.parseDouble(tokenizer.nextToken()), 
						Double.parseDouble(tokenizer.nextToken()), 
						Double.parseDouble(tokenizer.nextToken()));
				vertices.add(vertex);
				
				// Go to next line.
				line = reader.readLine();
				if (line != null)
					tokenizer = new StringTokenizer(line);
			}
			
			reader.close();
		}
	}
	
	/** Writes a set of model vertices to a file, where each line is a vertex "x y z".
	 *  
	 * @param filename	path to the file to save to.
	 * @param vertices	a list of the model vertices.
	 * 
	 * @throws IOException if output file could not be created.
	 */
	public static void writePointsToFile(String filename, ArrayList<Vertex3D> vertices) {
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(filename));
			for (int i = 0; i < vertices.size(); ++i)
				writer.write(String.format("%f %f %f\n", vertices.get(i).x, vertices.get(i).y, vertices.get(i).z));
			writer.close();
		} catch (IOException e) {
			System.err.println("Problem writing to file.");
		}
	}
	
	/** Encapsulates information of a Princeton Shape Benchmark styled model info file. */
	static class ModelInfo {
		Vertex3D center;
		double scale;
		double principalAxes[][];
		double principalVales[];
		public ModelInfo() {}
	}
	
}