package edu.txstate.mag262;
import java.util.ArrayList;

/** Performs Quadratic Form Distance calculations. */
public class QuadraticFormDistance {
	ArrayList<Vertex3D> form;
	double similarity[];
	
	@SuppressWarnings("unused")
	private QuadraticFormDistance() {}
	
	/** Constructs object with the information required for QFD calculations.
	 *  
	 * @param form		a set of points; there should be as many
	 *  points as there are entries in the vectors that will be 
	 *  compared. Each point represents the point in 3D space 
	 *  corresponding to a "dimension" of a vector; the "form".
	 *  
	 * @param theta		controls the shape of the matrix. Lower values
	 *  means elements are more similar, and vice-versa. Good values
	 *  to try are between 1 and 10.
	 */
	public QuadraticFormDistance(ArrayList<Vertex3D> form, double theta) {
		this.form = new ArrayList<>();
		for (Vertex3D p : form) this.form.add(new Vertex3D(p));
		calculateSimilarityMatrix(theta);
	}
	
	/** Calcuates matrix of similarities from bin i to bin j.
	 *  Calling this after initialization will recreate similarity matrix.
	 * 
	 * <pre>
	 * [1][x][x][x][x]
	 * [s][1][x][x][x]
	 * [s][s][1][x][x]
	 * [s][s][s][1][x]
	 * [s][s][s][s][1]</pre>
	 * 
	 * Does not save the x elements (to conserve memory).<br>
	 * Calculate index by: (((row + 1) * row) / 2) + col
	 * 
	 * @param theta		controls the shape of the matrix. Lower values
	 * means elements are more similar, and vice-versa. Good values
	 * to try are between 1 and 10.
	 */
	public void calculateSimilarityMatrix(double theta) {
		int dims = form.size();
		
		similarity = new double[dims*(dims+1)/2];
		
		// Similarity between a dimension and itself is 1.
		for (int i = 0; i < dims; ++i) similarity[(((i+2)*(i+1))/2)-1] = 1;
		
		// Rest of similarities.
		for (int i = 0; i < dims; ++i)
			for (int j = 0; j < i; ++j)
				similarity[(((i+1)*i)/2)+j] = Math.pow(Math.E, -theta*form.get(i).distanceTo(form.get(j)));
	}
	
	/** Calculates square distance between one vector to another. Vectors must have dimension equal to 
	 * the number of points in the form that was used to initialize this QFD object.
	 * 
	 * @param a		vector 'a'
	 * @param b		vector 'b'
	 * 
	 * @return QF distance from a to b.
	 */
	public double getDistance(double[] a, double[] b) {
		double result = 0;
		
		// Get differences between entries.
		double d[] = new double[form.size()];
		for (int i = 0; i < form.size(); ++i) d[i] = a[i]-b[i];
		
		// Basic Euclidian distance component.
		for (int i = 0; i < form.size(); ++i) result += d[i]*d[i];
		
		// The Quadratic Form Distance component.
		for (int i = 0; i < form.size(); ++i)
			for (int j = 0; j < i; ++j)
				result += 2*similarity[(((i+1)*i)/2)+j]*d[i]*d[j];
		
		return result;
	}
}
