package edu.txstate.mag262;
import java.util.ArrayList;

/** A spherical shape descriptor. Represented by a set of points, each at the center of a 
 * three-dimensional section of the descriptor. */
public class ShapeDescriptor {
	private ArrayList<Vertex3D> points;
	private SphereApproximator sphere;
	private double radius;
	private int shells;
	
	/** Creates a shape descriptor with the given information.
	 *  
	 *  @param radius	the radius of the spherical descriptor.
	 *  @param shells	the number of shells to use.
	 *  @param sphere	describes the sectors of the descriptor.
	 *  
	 *  @see SphereApproximator
	 *  */
	public ShapeDescriptor(double radius, int shells, SphereApproximator sphere) {
		this.sphere = sphere;
		this.radius = radius;
		this.shells = shells;
		calculateLocations();
	}
	
	/** Calculates the center locations of each descriptor section. */
	private void calculateLocations() {
		this.points = new ArrayList<>();
		
		double sep = radius / shells; // Calculate separation between shells.
		int sectors = sphere.numVerts();
		
		for (int i = 0; i < shells; ++i) {
			// Calculate the mid-shell radius.
			double msr = sep*(i + 0.5);
			
			// Get vertices at mid-shell radius and add to model.
			ArrayList<Vertex3D> temp = sphere.getVertsScaled(msr);
			for (int j = 0; j < sectors; ++j) points.add(temp.get(j));
		}
	}
	
	/** @return a deep copy of the descriptor points, centered at the origin. */
	public ArrayList<Vertex3D> getPoints() {
		ArrayList<Vertex3D> copy = new ArrayList<>();
		for (Vertex3D p : points) copy.add(new Vertex3D(p));
		return copy;
	}
	
	/** @return The radius of the descriptor. */
	public double getRadius() { return radius; }
	/** @return The number of shells of the descriptor. */
	public int getNumShells() { return shells; }
	/** @return The number of sectors of the descriptor. */
	public int getNumSectors() { return sphere.numVerts(); }
	/** @return The number of points, or sections, of the descriptor. */
	public int getNumPoints() { return points.size(); }
}
