package edu.txstate.mag262;
import java.util.ArrayList;

/** Approximates a sphere with points at the surface. */
public class SphereApproximator {
	private final double PHI = 1.6180339887498948482045868;
	
	private ArrayList<Vertex3D> verts;
	private ArrayList<Vertex3D> faces;
	
	private double radius = 1;
	
	@SuppressWarnings("unused")
	private SphereApproximator() {}
	
	/** Creates a unit sphere approximation using spherical coordinates.
	 * 
	 * @param horizontal	number of horizontal sectors (on xy-plane).
	 * @param vertical		number of vertical sectors. <b><i> Must be at least
	 * two! Three to count horizontal sectors also. </i></b>
	 */
	public SphereApproximator(int horizontal, int vertical) {
		--vertical;
		double sinc = Math.PI / vertical;
		double sazi = (Math.PI * 2.0) / horizontal;
		--vertical;
		
		verts = new ArrayList<>();
		verts.add(new Vertex3D(0, 0, 1));  // Zenith pos.
		verts.add(new Vertex3D(0, 0, -1)); // Zenith neg.
		
		double inc = 0;
		double azi = 0;
		
		// Generate points.
		for (int i = 0; i < vertical; ++i) {
			inc += sinc; // Next inclination angle.
			
			// Generate points around zenith.
			azi = 0;
			for (int a = 0; a < horizontal; ++a) {
				double lxy = Math.sin(inc);
				verts.add(new Vertex3D( lxy * Math.cos(azi),
									   lxy * Math.sin(azi),
									   Math.cos(inc)));
				azi += sazi;
			}
		}
	}
	
	/** Creates a sphere approximation by subdividing the vertices of an isosahedron. 
	 * The related characteristic equation is V - 0.5 * F = 2. Number of edges can be 
	 * related to number of faces as such: E = (F * 3) / 2 = 1.5 * F.
	 *  
	 * @param order		number of subdivisions, higher number means more vertices.
	 * <br><i>In particular, <b>the number of vertices is (2 + 10 * 4^order)</b></i>.
	 */
	public SphereApproximator(int order) {
		
		this.radius = 1.902113032590307;
		
		// Create initial isosahedron.
		verts = new ArrayList<>();
		verts.add(new Vertex3D(0, 1, PHI));
		verts.add(new Vertex3D(0, 1, -PHI));
		verts.add(new Vertex3D(0, -1, PHI));
		verts.add(new Vertex3D(0, -1, -PHI));
		verts.add(new Vertex3D(1, PHI, 0));
		verts.add(new Vertex3D(1, -PHI, 0));
		verts.add(new Vertex3D(-1, PHI, 0));
		verts.add(new Vertex3D(-1, -PHI, 0));
		verts.add(new Vertex3D(PHI, 0, 1));
		verts.add(new Vertex3D(PHI, 0, -1));
		verts.add(new Vertex3D(-PHI, 0, 1));
		verts.add(new Vertex3D(-PHI, 0, -1));
		
		// Specify the initial faces.
		faces = new ArrayList<>();
		faces.add(new Vertex3D(0, 4, 8));
		faces.add(new Vertex3D(0, 4, 6));
		faces.add(new Vertex3D(0, 6, 10));
		faces.add(new Vertex3D(0, 2, 10));
		faces.add(new Vertex3D(0, 8, 2));
		faces.add(new Vertex3D(1, 4, 6));
		faces.add(new Vertex3D(1, 4, 9));
		faces.add(new Vertex3D(1, 3, 9));
		faces.add(new Vertex3D(1, 3, 11));
		faces.add(new Vertex3D(1, 6, 11));
		faces.add(new Vertex3D(2, 5, 8));
		faces.add(new Vertex3D(2, 5, 7));
		faces.add(new Vertex3D(2, 7, 10));
		faces.add(new Vertex3D(3, 5, 9));
		faces.add(new Vertex3D(3, 5, 7));
		faces.add(new Vertex3D(3, 7, 11));
		faces.add(new Vertex3D(4, 8, 9));
		faces.add(new Vertex3D(5, 8, 9));
		faces.add(new Vertex3D(6, 10, 11));
		faces.add(new Vertex3D(7, 10, 11));
		
		for (int i = 0; i < order; ++i) subdivide();
	}
	
	/** Subdivides the vertices of the current sphere to create more vertices.	 */
	private void subdivide() {
		ArrayList<Vertex3D> newFaces = new ArrayList<>();
		for (int i = 0; i < faces.size(); ++i) {
			Vertex3D f = faces.get(i);
			
			// Calculate midpoints.
			Vertex3D a = verts.get((int)f.x).midpoint(verts.get((int)f.y));
			Vertex3D b = verts.get((int)f.y).midpoint(verts.get((int)f.z));
			Vertex3D c = verts.get((int)f.z).midpoint(verts.get((int)f.x));
			
			// Bring them to the surface of the 'sphere'
			a.scaleToLength(radius);
			b.scaleToLength(radius);
			c.scaleToLength(radius);
			
			// If a midpoint is already a vertex, use that index.
			// If not, add it to the list.
			int ai = vertContains(a);
			if (ai == -1) {
				ai = verts.size();
				verts.add(a);
			}
			int bi = vertContains(b);
			if (bi == -1) {
				bi = verts.size();
				verts.add(b);
			}
			int ci = vertContains(c);
			if (ci == -1) {
				ci = verts.size();
				verts.add(c);
			}
			
			// Create new faces.
			newFaces.add(new Vertex3D(ai, f.x, ci));
			newFaces.add(new Vertex3D(f.y, ai, bi));
			newFaces.add(new Vertex3D(bi, ci, f.z));
			newFaces.add(new Vertex3D(ai, ci, bi));
		}
		faces = newFaces;
	}
	
	/** Search the list of vertices for a particular vertex.
	 * 
	 * @param v		the vertex to look for (sees if values match, not if it's the same object).
	 * @return The index if found, -1 otherwise.
	 */
	private int vertContains(Vertex3D v) {
		for (int i = 0; i < verts.size(); ++i)
			if (v.equals(verts.get(i))) return i;
		return -1;
	}
	
	/** @param radius	the radius to scale the sphere to.
	 *  @return The points of the sphere approximation, scaled to the new radius. 
	 *  */
	public ArrayList<Vertex3D> getVertsScaled(double radius) {
		radius = Math.abs(radius);
		ArrayList<Vertex3D> newVerts = new ArrayList<Vertex3D>();
		for (int i = 0; i < verts.size(); ++i) {
			Vertex3D v = new Vertex3D(verts.get(i));
			v.scaleToLength(radius);
			newVerts.add(v);
		}
		return newVerts;
	}
	
	/** @return The radius of the approximated sphere. */
	public double getRadius() { return radius; }
	
	/** @return A deep copy of the sphere vertices. */
	public ArrayList<Vertex3D> getVerts() {
		ArrayList<Vertex3D> copy = new ArrayList<>();
		for (Vertex3D v : verts) copy.add(new Vertex3D(v));
		return copy; 
	}
	
	/** @return A deep copy of the sphere faces, or (null) if faces were not used. */
	public ArrayList<Vertex3D> getFaces() {
		if (faces == null) return null;
		ArrayList<Vertex3D> copy = new ArrayList<>();
		for (Vertex3D f : faces) copy.add(new Vertex3D(f));
		return copy;
	}
	
	/** @return The number of vertices. */
	public int numVerts() { return verts.size(); }
	/** @return The number of faces. */
	public int numFaces() { return faces == null ? 0 : faces.size(); }
}
