package edu.txstate.mag262;

/** A vertex in 3D space with vector operations. */
public class Vertex3D {

	public double x, y, z;
	private final static double ABS_EPS = 0.00000000000000011102230246251565;
	// Note: may need to calculate epsilon per machine.
	
	/** Creates a new Vertex3D object. */
	public Vertex3D(double x, double y, double z) {
		this.x = x;
		this.y = y;
		this.z = z;
	}
	
	/** Converts an 3D double array to a Vertex3D object. */
	public Vertex3D(double a[]) {
		this(a[0], a[1], a[2]);
	}
	
	/** Copy constuctor. */
	public Vertex3D(Vertex3D v) {
		this.x = v.x;
		this.y = v.y;
		this.z = v.z;
	}
	
	/** Elementwise addition of another vertex to this one.
	 * @param v	other vertex. */
	public void add(Vertex3D v) {
		this.x += v.x;
		this.y += v.y;
		this.z += v.z;
	}
	
	/** Elementwise subtraction of another vertex from this one. 
	 * @param v	other vertex. */
	public void sub(Vertex3D v) {
		this.x -= v.x;
		this.y -= v.y;
		this.z -= v.z;
	}
	
	/** Elementwise multiplication with a scalar. 
	 * @param s	the scalar value. */
	public void times(double s) {
		this.x *= s;
		this.y *= s;
		this.z *= s;
	}
	
	/** Elemenwise comparison of equality between this vertex and another.
	 * @param v	 other vertex.
	 * @return Result of comparison. */
	public boolean equals(Vertex3D v) {
		return (equalEps(x, v.x) && equalEps(y, v.y) && equalEps(z, v.z));
	}
	
	/** Compares two doubles using absolute and relative
	 *  epsilon checking. Algorithm by Bruce Dawson, found
	 *  <a href="randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/">
	 *  here</a>.
	 *  
	 *  @param a	double 'a'
	 *  @param b	double 'b'
	 *  
	 *  @return Result of comparison.
	 *  */
	public static boolean equalEps(double a, double b) {
		// Difference less than epsilon? Zero-checking.
		double diff = Math.abs(b - a);
		if (diff <= ABS_EPS) return true;
		
		// Numbers have different signs?
		if ((a < 0) == (b >= 0)) return false;
		if ((b < 0) == (a >= 0)) return false;
		
		// Relative epsilon check.
		double largest = Math.max(a, b);
		if (diff <= largest * ABS_EPS * 1.8) return true;
		
		return false;
	}
	
	/** Calculates the midpoint between this vertex and another.
	 * @param v	 other vertex.
	 * @return The midpoint as another Vertex3D object. */
	public Vertex3D midpoint(Vertex3D v) {
		return new Vertex3D((v.x + x) / 2.0, (v.y + y) / 2.0, (v.z + z) / 2.0);
	}
	
	/** The squared distance from this vertex to another.
	 * Faster calculation than the regular distance. Useful
	 * if you just need to check if one distance is larger
	 * than another. 
	 * 
	 * @param v	other vertex. 
	 * @return The squared Euclidian distance to the other vertex */
	public double sqDistanceTo(Vertex3D v) {
		double xdiff = v.x - x;
		double ydiff = v.y - y;
		double zdiff = v.z - z;
		return ((xdiff*xdiff) + (ydiff*ydiff) + (zdiff*zdiff));
	}
	
	/** The distance between this vertex and another.
	 * @param v other vertex.
	 * @return The Euclidian distance to the other vertex. */
	public double distanceTo(Vertex3D v) {
		return Math.sqrt(sqDistanceTo(v));
	}
	
	/** Calculates the squared Euclidean distance between two 
	 * vectors of arbitrary length.
	 *  
	 * @param a		vector a
	 * @param b		vector b
	 * @return The squared Euclidian distance between a and b. 
	 */
	public static double sqDistance(double[] a, double[] b) {
		double result = 0.0;
		for (int i = 0; i < a.length; ++i) {
			double diff = b[i] - a[i];
			result += diff*diff;
		}
		return result;
	}
	
	/** @return The length of the vertex as a vector. */
	public double length() {
		return Math.sqrt(x*x + y*y + z*z);
	}
	
	/** Scales this vertex as a vector.
	 * @param s  the length to scale to. */
	public void scaleToLength(double s) {
		double t = length() / s;
		this.x /= t;
		this.y /= t;
		this.z /= t;
	}
	
	/** @return The Vertex3D object as a double array. */
	public double[] asArray() {
		return new double[] {x, y, z};
	}
	
	@Override
	public String toString() {
		return String.format("{%.6f, %.6f, %.6f}", x, y, z);
	}
}
